BeginPackage["HighEnergyPhysics`Complex`"];

END::usage = "END indicates to ComplexTimes and ComplexPlus that it
has reached the end of a sequence of complex numbers."

SQR::usage = "SQR[expr] is used the C program is indicate expr^2."

ComplexQ::usage = "ComplexQ[expr] returns True if expr contains any complex
quantity, and False otherwise."

ComplexForm::usage = "ComplexForm[expr] prints a version of expr that
can be used in a C program with the package complex.c."

SetAttributes[ComplexForm, Listable]

ConstructComplex::usage = "ConstructComplex[x, y] represents the
complex number x+ I y in a C program."

DoubleToComplex::usage = "DoubleToComplex[x] converts a real number
into a complex number in a C program."

ComplexPlus::usage = "ComplexPlus[x1, x2, ..., xn, END] adds together
the complex numbers x1, x2, ..., xn in the C program."

ComplexTimes::usage = "ComplexTimes[x1, x2, ..., xn, END]
multiplies together the complex numbers x1, x2, ..., xn in the C
program."

TimesByDouble::usage = "TimesByDouble[x, y] mutiplies the complex
number x by the real number y in the C program."

ComplexPower::usage = "ComplexPower[z1, z2] represents z1^z2 in the C program where z1 and z2 are complex numbers."

(* *************************************************************** *)

Begin["Complex`Private`"]


(* *************************************************************** *)


(*  ComplexQ definitions  *)

ComplexQ[] = False

ComplexQ[_] = False

ComplexQ[_Symbol] = False

ComplexQ[_Integer] = False

ComplexQ[_Real] = False

ComplexQ[_Complex] = True

ComplexQ[Power[expr_, _]] := ComplexQ[expr]

ComplexQ[expr_Times] := Or @@ (ComplexQ /@ (List @@ expr))

ComplexQ[expr_Plus] := Or @@ (ComplexQ /@ (List @@ expr))

ComplexQ[expr_List] := Or @@ (ComplexQ /@ expr)

ComplexQ[Sqrt[expr_]] := ComplexQ[expr]

ComplexQ[_ConstructComplex] = True

ComplexQ[ConstructComplex] ^= True

ComplexQ[_DoubleToComplex] = True

ComplexQ[DoubleToComplex] ^= True

ComplexQ[_ComplexPlus] = True

ComplexQ[ComplexPlus] ^= True

ComplexQ[_ComplexTimes] = True

ComplexQ[ComplexTimes] ^= True

ComplexQ[_TimesByDouble] = True

ComplexQ[TimesByDouble] ^= True

(* *************************************************************** *)

(*  ComplexForm  *)

ComplexForm[expr_] := expr //. complexformrules

complexformrules = {

Complex[x_, y_] :> ConstructComplex[x, y],

Literal[Plus[expr__]] :> Append[
	ComplexPlus @@ (If[!ComplexQ[#], DoubleToComplex[#], #]& /@ {expr}), 
			END] /; ComplexQ[{expr}],

expr__Times :> 
	Block[{comp, noncomp},
	comp  = Select[List @@ expr, ComplexQ];
	noncomp = Complement[List @@ expr, comp];
	If[Length[comp] == 1,
	TimesByDouble[First[comp], Times @@ noncomp],
		If[noncomp === {},
		Append[Release[ComplexTimes @@ comp], END],
		TimesByDouble[Append[Release[ComplexTimes @@ comp], END],
			Times @@ noncomp]]]
	] /; ComplexQ[expr],

Power[x_, 2] :> SQR[x] /; !ComplexQ[x],

Power[x_, n_Integer?Positive] :> ComplexTimes @@ 
			(Append[Table[x, Release[{n}]], END]) /;
				ComplexQ[x],

Power[x_, y_] :> ComplexPower[x, ConstructComplex[y, 0]] /;
				ComplexQ[x] && !ComplexQ[y]

}
	
(* *************************************************************** *)



End[];

EndPackage[]

