BeginPackage["HighEnergyPhysics`SpinorTechniques`",
"HighEnergyPhysics`CrossSection`",
"HighEnergyPhysics`RelativisticKinematics`", "HighEnergyPhysics`Spinor`",
"HighEnergyPhysics`NonCommutativeMultiply`", "HighEnergyPhysics`Complex`"]

PrepareMatrixElementForCProgram::usage =
"PrepareMatrixElementForCProgram[choppedmatrixelement, polarizations,
contractions, simplifyingrules, collect, momenta, outputfile] will
convert your matrixelement (which is choppedmatrixelement *
polarizations) into spinor-technique form as well as doing some
simplifications for you.  contractions are the contractions that need
to be done except for those in the polarizations.  simplifyingrules
are rules you would like applied to the matrix-element.  collect lists
the variables that can be factored out of the matrix-element,
typically coupling constants.  momenta is a list containing first a
list of incoming momenta followed by a list of outgoing momenta.
outputfile is the file you want the prepared matrix-element saved in
as well as some C-statements for ReplacedExpression's and
DotProduct's."

Options[PrepareMatrixElementForCProgram] = {Mandelstam -> {}}

DoubleReplacedExpression::usage = "DoubleReplacedExpression[x]
represents a real expression that was substituted when
PrepareMatrixElementForCProgram was executed."

ComplexReplacedExpression::usage = "ComplexReplacedExpression[x]
represents a complex expression that was substituted when
PrepareMatrixElementForCProgram was executed."

ReplacedExpressionRule::usage = "ReplacedExpressionRule[x]'s contain
the rules that associate ComplexReplacedExpression[x]'s and
DoubleReplacedExpression[x]'s with the expressions they represent."

LightlikeVectorDecayedFrom::usage = "LightlikeVectorDecayedFrom[p, n]
represents a lightlike vector used for the polarizations of massive
vector bosons.  It obeys p = LightlikeVectorDecayedFrom[p,1] +
LightlikeVectorDecayedFrom[p,2] where p is the momentum of the
external massive vector boson."

MaximumTolerableDepth::usage = "MaximumTolerableDepth indicates what
depth expressions should be before they are broken down before
splicing."

MasslessInAndOutList::usage = "MasslessInAndOutList is a variable that
is given the list of spinors and massless vector boson polarizations
after PolarizationCombinations is called."

Opposite::usage = "Opposite[p] represents a four-momentum which has
the same energy but opposite three momentum to the four-momentum p."

(* 
ChooseLightlikeVectorNotCollinearWith::usage =
"ChooseLightlikeVectorNotCollinearWith[PhaseSpaceIntegral[...]] trys
it's best to intelligently give values to
LightlikeVectorNotCollinearWith[...] inside the integrand.
*)

Squared::usage = "Squared[me] indicates to
CEvaluatePhaseSpaceIntegrate that the matrix-element needs to be
squared."

SetAttributes[Squared, Listable]

CEvaluatePhaseSpaceIntegrate::usage =
"CEvaluatePhaseSpaceIntegrate[PhaseSpaceIntegrate[...]] will prepare a
C program which will numerically calculate your cross-section or
decay...  isn't that great?!!  It will automatically evaluate any
LightlikeVectorNotCollinearWith present in the matrix-element
squared."

Options[CEvaluatePhaseSpaceIntegrate] = {SumOverPolarizations ->
False,	InputFile -> "skel.mc", OutputFile -> "", IntegrandName ->
"integrand", DoFactor -> False, CouplingConstant -> {}}

SumOverPolarizations::usage = "SumOverPolarizations is an option for
CEvaluatePhaseSpaceIntegrate."

InputFile::usage = "InputFile is an option for CEvaluatePhaseSpaceIntegrate."

OutputFile::usage = "OutputFile is an option for CEvaluatePhaseSpaceIntegrate."

IntegrandName::usage = "IntegrandName is an option for
CEvaluatePhaseSpaceIntegrate."

DoFactor::usage = "DoFactor is an option for CEvaluatePhaseSpaceIntegrate."

CouplingConstant::usage = "CouplingConstant is an option for
CEvaluatePhaseSpaceIntegrate.  It lists the factors which you would
like to be factored out in the factoring process.  For example you
might include Alpha and Pi.  If the option CouplingConstant is used it
turns on the DoFactor option automatically."

PolarizationCombinations::usage =
"PolarizationCombinations[me] returns a list of the possible
polarization combinations of the initial and final particles for the
matrix-element me over the polarizations which are not explicitly
written in the matrix-element."

LightlikeVectorNotCollinearWith::usage = "LightlikeVectorNotCollinearWith[p]
represents a lightlike vector that is not collinear with p."

SpinorS::usage = "SpinorS[p, q] is the same as
SpinorUbar[p, Right]**SpinorU[q, Left]. Both p and q have to be
massless."

SpinorT::usage = "SpinorT[p, q] is the same as
SpinorUbar[p, Left]**SpinorU[q, Right]. Both p and q have to be
massless."

SpinorST::usage = "SpinorST[pol1, p1, p2, ..., pn, pol2] is the same as
SpinorUbar[p1, pol1]**Slash[p2]**...**SpinorU[pn, pol2].
All the pi have to be massless."

ConvertToST::usage = "ConvertToST[expr] attempts to convert DiracGamma5
and HelicityProjection occurrences in expr to SpinorS and SpinorT
objects."

Options[ConvertToST] = {ConvertToMassless -> {}}

ConvertToMassless::usage = "ConvertToMassless is an option for ConvertToST."

STToTraces::usage = "STToTraces[expr] attempts to convert products of
SpinorS's and SpinorT's in expr to traces using the rule
SpinorS[p1,p2]*SpinorT[p2, p3]...SpinorS[pn, p1] =
Trace[Slash[p1]**Slash[p2]**...**Slash[pn]**HelicityProjection[Right]]"

SetAttributes[STToTraces, Listable]



(* ************************************************************* *)


Begin["SpinorTechniques`Private`"]

Clear[tr, chooseLightlikeVectorNotCollinearWith, cEvaluatePhaseSpaceIntegrate1]

(* ************************************************************* *)

Combinations[l_List, k_Integer] := Flatten[
             Block[{i, ll=Length[l],
             m = Table[Unique["t"], {k}]},
             Table[l[[Release[m]]],
             Release[{m[[1]], ll-k+1}],
             Release[Sequence @@ Table[{m[[i]], m[[i-1]]+1, ll-k+i}, {i,
             2, k}]]]], k-1]

LeftRightToNumber[Left] = -1
LeftRightToNumber[Right] = 1

ReverseInPlace[p___] := Sequence @@ Reverse[{p}]

(* ************************************************************ *)

(*  PolarizationCombinations  *)

PolarizationCombinations[me_] := Block[
	{tmp = Level[me, {0,Infinity}], u, ubar, v, vbar, mless, rules1,
	rules2, rules},
	u = Union[Cases[tmp, SpinorU[_]]];
	ubar = Union[Cases[tmp, SpinorUbar[_]]];
	v = Union[Cases[tmp, SpinorV[_]]];
	vbar = Union[Cases[tmp, SpinorVbar[_]]];
	mless = Union[Cases[tmp, MasslessVectorPolarization[_,_]]];

	MasslessInAndOutList = Join[u, ubar, v, vbar, mless];

	If[MasslessInAndOutList =!= {},
	rules1 = {Insert[#, Left, -1], Insert[#, Right, -1]}& /@ 
		MasslessInAndOutList;
	rules1 = Flatten[Outer[List, Sequence @@ rules1],
		Length[MasslessInAndOutList]-1];
	rules1 = Apply[Rule,
		Transpose[{MasslessInAndOutList, #}]& /@ rules1, {2}];
	rules2 = {Insert[#, Right, -1], Insert[#, Left, -1]}& /@ 
		MasslessInAndOutList;
	rules2 = Flatten[Outer[List, Sequence @@ rules2],
		Length[MasslessInAndOutList]-1];
	rules2 = Apply[Rule,
		Transpose[{#, Table[0, {Length[MasslessInAndOutList]}]}]& /@ 
		rules2, {2}];
	rules = Flatten /@ Transpose[{rules1, rules2}];
	(me /. #)& /@ rules,
	me]]

(* ************************************************************** *)

(*  PhysicsForm1  *)

(*  these won't work because they go AFTER the /@ rules for PhysicsForm1 *)

Spinor`Private`PhysicsForm1[
	MasslessVectorPolarization[p_, mu_, pol___]] :=
	SequenceForm["eps",ColumnForm[{mu," ",
	Spinor`Private`PhysicsForm1[pol]}, Left, Center], "[", p, "]"]

Spinor`Private`PhysicsForm1[
	HeavyVectorPolarization[p_, mu_, pol___]] :=
	SequenceForm["eps",ColumnForm[{mu," ",
	Spinor`Private`PhysicsForm1[pol]}, Left, Center], "[", p, "]"]

Spinor`Private`PhysicsForm1[SpinorS[p_, q_]] := 
    	SequenceForm["s[", p, ",", q, "]"]

Spinor`Private`PhysicsForm1[SpinorT[p_, q_]] := 
    	SequenceForm["t[", p, ",", q, "]"]

(* ************************************************************* *)

(* Squared *)

Squared[0] = 0

Squared[] = 1

(*
Squared[Times[x_?(UnsameQ[Head[#], Plus]&), y__]] := AbsSquared[x] Squared[y]
*)
(*
Squared[x_?((UnsameQ[Head[#], Plus] && UnsameQ[Head[#], Times])&)] :=
	AbsSquared[x]
*)


(* ************************************************************* *)

(* Complex rules *)

ComplexQ[_SpinorS] = True

ComplexQ[_SpinorT] = True

ComplexQ[_DotProduct] = False

ComplexQ[_Squared] = False


(* ************************************************************* *)

(* Conjugation rules *)


Unprotect[Abs]
Abs/: Conjugate[x_Abs] := x
Protect[Abs]

SpinorS/: Conjugate[Literal[SpinorS[p1_, p2_]]] := -SpinorT[p1, p2]

SpinorT/: Conjugate[Literal[SpinorT[p1_, p2_]]] := -SpinorS[p1, p2]

LightlikeVectorNotCollinearWith/: Conjugate[
	LightlikeVectorNotCollinearWith[p_]] := 
		LightlikeVectorNotCollinearWith[p]

LightlikeVectorDecayedFrom/: Conjugate[
	LightlikeVectorDecayedFrom[p_, n_]] := LightlikeVectorDecayedFrom[p,n]


(* ************************************************************* *)

(*  SpinorS & SpinorT rules  *)

Literal[SpinorS[p_, p_]] = 0

Literal[SpinorT[p_, p_]] = 0

Literal[SpinorST[__, p_, p_, __]] = 0

Literal[SpinorST[pol1_, p1_, p___, pn_, pol2_]] := 0 /;
		(pol1 == pol2 && EvenQ[Length[{p}]]) ||
		(pol1 == -pol2 && OddQ[Length[{p}]])

Literal[SpinorS[p1_, p2_]] := -SpinorS[p2, p1] /; !OrderedQ[{p1, p2}]

Literal[SpinorT[p1_, p2_]] := -SpinorT[p2, p1] /; !OrderedQ[{p1, p2}]


SpinorS/: Literal[SpinorS[p1_, p2_]] * Literal[SpinorT[p1_, p2_]] := 
    -2 DotProduct[p1, p2]

SpinorS/: SpinorS[p1_, p2_] * Power[SpinorT[p1_, p2_], n_] :=
	-2 DotProduct[p1,p2] SpinorT[p1, p2]^(n-1) /; n > 1

SpinorT/: SpinorT[p1_, p2_] * Power[SpinorS[p1_, p2_], n_] :=
	-2 DotProduct[p1,p2] SpinorS[p1, p2]^(n-1) /; n > 1


SpinorT[p1:{_, _, _, _}, p2:{_, _, _, _}] := Conjugate[SpinorS[p2, p1]]

SpinorS[{px_, py_, pz_, pt_}, {kx_, ky_, kz_, kt_}] :=
	(py + I pz) Sqrt[(kt - kx)/(pt - px)] - 
	(ky + I kz) Sqrt[(pt - px)/(kt - kx)]

(*
SpinorS[{px_, py_, pz_, pt_}, {kx_, ky_, kz_, kt_}] :=
  Block[{pb = Sqrt[py^2+pz^2], kb = Sqrt[ky^2+kz^2]},
Sqrt[(kt+kx)(pt-px)] If[Chop[pb]===0, 1, (py - I pz)/pb] -
Sqrt[(pt+px)(kt-kx)] If[Chop[kb]===0, 1, (ky - I kz)/kb]]
*)

(*  I don't know why the next statement is here *)
(*
SpinorS[{px_, py_, pz_, pt_}, {kx_, ky_, kz_, kt_}] :=
	      Block[{pplus = px+pt, pminus = pt-px,
qplus = kt+kx, qminus = kt-kx, ppb = py-I pz, pbot, qqb = ky-I kz,
qbot},
pbot = Abs[ppb]; ppb = If[Chop[pbot] === 0, 1, ppb / pbot];
qbot = Abs[qqb]; qqb = If[Chop[qbot] === 0, 1, qqb / qbot];
ppb Sqrt[qplus pminus] - qqb Sqrt[pplus qminus]]
*)

SpinorST/: Literal[SpinorST[pol1_, p___, pn_, pol2_] *
		SpinorST[pol2_, pn_, pp___, pol3_]] :=
		SpinorST[pol1, p, pn, pp, pol3]

SpinorST/: Literal[SpinorST[pol1_, p1_, p___, pol2_] *
		SpinorST[pol3_, p1_, pp___, pol4_]] :=
   		LeftRightToNumber[pol1] LeftRightToNumber[pol2] *
		SpinorST[-pol2, ReverseInPlace[p], p1, pp, pol4] /; 
							pol1 == -pol3

SpinorST/: Literal[SpinorST[pol1_, p___, p1_, pol2_] *
		SpinorST[pol3_, pp___, p1_, pol4_]] :=
   		LeftRightToNumber[pol3] LeftRightToNumber[pol4] *
		SpinorST[pol1, p, p1, ReverseInPlace[pp], -pol3] /;
							pol4 == -pol2

SpinorST/: Literal[SpinorST[_, p1_, p___, pn_, pol_]] *
		Literal[Conjugate[SpinorST[_, p1_, p___, pn_, pol_]]] :=
		Trace[HelicityProjection[(-1)^Length[{p}] pol] ** 
           	(NonCommutativeMultiply @@ 
	    	(Slash /@ {p, pn, ReverseInPlace[p], p1}))]

SpinorST/: SpinorST[pol_, p1_, p___, p1_, pol_] :=
  		Trace[HelicityProjection[pol] ** (NonCommutativeMultiply @@
		(Slash /@ {p1, p}))]



(*
SpinorST/: Spinor`Private`AbsSquared1[SpinorST[pol_, p1_, p___, pn_, pol_]] = 
    Trace[HelicityProjection[(-1)^Length[{p}] (-pol)] ** 
           (NonCommutativeMultiply @@ 
	    (Slash /@ {p, pn, ReverseInPlace[p], p1}))]
*)

(* ************************************************************ *)

(*  Converting S,T expr to Traces  *)

STToTraces[x_Plus] := STToTraces /@ x

STToTraces[a_ b_] := a STToTraces[b] /; 
            FreeQ[a, SpinorS] && FreeQ[a, SpinorT] && FreeQ[a, SpinorST]

STToTraces[expr_] := expr //.
{
(* the below rule is temporary until I find something better *)
(* the rule also appears in ConvertToST *)
SpinorS[p1_, p2_]^n1_. * SpinorT[p1_, p2_]^n2_. :>
	(-2 DotProduct[p1,p2])^n1 SpinorT[p1, p2]^(n2-n1) /; (n1>0 && n2>=n1),

SpinorS[p1_, p2_]^n1_. * SpinorT[p1_, p2_]^n2_. :>
	(-2 DotProduct[p1,p2])^n2 SpinorS[p1, p2]^(n1-n2) /; (n2>0 && n1>=n2),

SpinorS[p1_, p2_] :> SpinorST[Right, p1, p2, Left],

SpinorT[p1_, p2_] :> SpinorST[Left, p1, p2, Right],

SpinorST[pol1_, p1___, p2_, pol2_]^n1_. *
	SpinorST[pol2_, p2_, p3___, pol3_]^n2_. :>
	SpinorST[pol1, p1, p2, p3, pol3]^n1 * 
	SpinorST[pol2, p2, p3, pol3]^(n2-n1) /; (n1>0 && n2>=n1),

SpinorST[pol1_, p1___, p2_, pol2_]^n1_. *
	SpinorST[pol2_, p2_, p3___, pol3_]^n2_. :>
	SpinorST[pol1, p1, p2, p3, pol3]^n2 * 
	SpinorST[pol1, p1, p2, pol2]^(n1-n2) /; (n2>0 && n1>=n2),


SpinorST[pol1_, p1_, p___, pol2_]^n1_. *
	SpinorST[pol3_, p1_, pp___, pol4_]^n2_. :>
   	(LeftRightToNumber[pol1] LeftRightToNumber[pol2] *
	SpinorST[-pol2, ReverseInPlace[p], p1, pp, pol4])^n1 *
	SpinorST[pol3, p1, pp, pol4]^(n2-n1) /;
	(pol1 == -pol3 && n1>0 && n2>=n1),

SpinorST[pol1_, p1_, p___, pol2_]^n1_. *
	SpinorST[pol3_, p1_, pp___, pol4_]^n2_. :>
   	(LeftRightToNumber[pol1] LeftRightToNumber[pol2] *
	SpinorST[-pol2, ReverseInPlace[p], p1, pp, pol4])^n2 *
	SpinorST[pol1, p1, p, pol2]^(n1-n2) /;
	(pol1 == -pol3 && n2>0 && n1>=n2),

SpinorST[pol1_, p___, p1_, pol2_]^n1_. *
	SpinorST[pol3_, pp___, p1_, pol4_]^n2_. :>
   	(LeftRightToNumber[pol3] LeftRightToNumber[pol4] *
	SpinorST[pol1, p, p1, ReverseInPlace[pp], -pol3])^n1 *
	SpinorST[pol3, pp, p1, pol4]^(n2-n1) /;
	(pol4 == -pol2 && n1>0 && n2>=n1),

SpinorST[pol1_, p___, p1_, pol2_]^n1_. *
	SpinorST[pol3_, pp___, p1_, pol4_]^n2_. :>
   	(LeftRightToNumber[pol3] LeftRightToNumber[pol4] *
	SpinorST[pol1, p, p1, ReverseInPlace[pp], -pol3])^n2 *
	SpinorST[pol1, p, p1, pol2]^(n1-n2) /;
	(pol4 == -pol2 && n2>0 && n1>=n2),

SpinorST[pol1_, p1_, p___, pn_, pol2_]^n1_. *
	Conjugate[SpinorST[pol1_, p1_, p___, pn_, pol2_]]^n2_. :>
	(Trace[HelicityProjection[(-1)^Length[{p}] pol2] ** 
        (NonCommutativeMultiply @@
	(Slash /@ {p, pn, ReverseInPlace[p], p1}))])^n1 *
	Conjugate[SpinorST[pol1, p1, p, pn, pol2]]^(n2-n1) /; (n1>0 && n2>=n1),

SpinorST[pol1_, p1_, p___, pn_, pol2_]^n1_. *
	Conjugate[SpinorST[pol1_, p1_, p___, pn_, pol2_]]^n2_. :>
	(Trace[HelicityProjection[(-1)^Length[{p}] pol2] ** 
        (NonCommutativeMultiply @@
	(Slash /@ {p, pn, ReverseInPlace[p], p1}))])^n2 *
	SpinorST[pol1, p1, p, pn, pol2]^(n1-n2) /; (n2>0 && n1>=n2)

}


(* ************************************************************ *)

(* Converting to ST expression *)

LightlikeVectorNotCollinearWith/:  Mass[_LightlikeVectorNotCollinearWith] = 0

LightlikeVectorDecayedFrom/:  Mass[_LightlikeVectorDecayedFrom] = 0

LightlikeVectorDecayedFrom/:
	DotProduct[LightlikeVectorDecayedFrom[p_, 1], 
		LightlikeVectorDecayedFrom[p_, 2]] := Mass[p]^2/2

ConvertToST[expr_, opts___] := Block[
	{x = ConvertToMassless /. {opts} /. Options[ConvertToST]},
	x = Switch[x,
		{}, {},
		{_Rule..},
			Flatten[Block[{a = #[[1]], b = #[[2]]},
			{Slash[a] -> ExpandSlash[Slash[b]],
			a[mu_] :> Release[b[mu]]}]& /@ x]];
	ConvertToST1[Expand[ExpandSlash[expr]], x]]


ConvertToST1[a_Plus, rules_] := ConvertToST1[#, rules]& /@ a

ConvertToST1[a_, rules_] := Block[
	{tmp = a /. rules //. ConvertToSTRules},
	If[tmp === a,  tmp,  ConvertToST1[Expand[ExpandSlash[tmp]], rules]]]

ConvertToSTRules = {
	SpinorV[p_, pol_] :> SpinorU[p, pol] /; Mass[p] == 0,

	SpinorVbar[p_, pol_] :> SpinorUbar[p, pol] /; Mass[p] == 0,

	Literal[NonCommutativeMultiply[a___, b_Plus, c___]] :>
		(NonCommutativeMultiply[a, #, c]&) /@ b,

	Literal[NonCommutativeMultiply[
	___, SpinorUbar[_, pol_],SpinorU[_, pol_], ___]] :> 0,

	Literal[ncm:NonCommutativeMultiply[___, DiracGamma[mu_], ___]] *
	Literal[pp:Plus[(_[___, mu_,___])..]] :> Expand[ncm pp],

	Literal[NonCommutativeMultiply[a___, DiracGamma[mu_], b___]] *
	p_[mu_] :> NonCommutativeMultiply[a, Slash[p], b],

	Literal[NonCommutativeMultiply[a___, DiracGamma[mu_], b___]] *
	G[mu_, nu_] :> NonCommutativeMultiply[a, DiracGamma[nu], b],

	Literal[NonCommutativeMultiply[a___, Slash[p_], b___]] :>
		NonCommutativeMultiply[
		a, SpinorU[p, Left], SpinorUbar[p, Left], b] +
		NonCommutativeMultiply[
		a, SpinorU[p, Right], SpinorUbar[p, Right], b] /;Mass[p] === 0,

	Literal[NonCommutativeMultiply[a___, Slash[p_], b___]] :>
		NonCommutativeMultiply[a,
		Slash[LightlikeVectorDecayedFrom[p, 1]] +
		Slash[LightlikeVectorDecayedFrom[p, 2]], b] /; Mass[p] =!= 0,

	MasslessVectorPolarization[p_, mu_, pol_] :> 
		Block[{tmp = LightlikeVectorNotCollinearWith[p]},
		SpinorUbar[p, pol] ** DiracGamma[mu] ** SpinorU[tmp, pol] / 
		(2 Sqrt[Abs[DotProduct[p, tmp]]])],

	HeavyVectorPolarization[p_, mu_] :>
		SpinorUbar[LightlikeVectorDecayedFrom[p, 1], Left] **
		DiracGamma[mu] **
		SpinorU[LightlikeVectorDecayedFrom[p, 2], Left] *
		Sqrt[3/(2 Pi)]/(2 Mass[p]),

(*  I think the next rule is only valid for massless p1 & p2  *)
	Literal[NonCommutativeMultiply[a___, DiracGamma[mu_], b___]] *
	Literal[NonCommutativeMultiply[
	SpinorUbar[p1_, pol_], DiracGamma[mu_], SpinorU[p2_, pol_]]] :>
		2 (NonCommutativeMultiply[
		a, SpinorU[p2, pol], SpinorUbar[p1, pol], b] +
		NonCommutativeMultiply[
		a, SpinorU[p1, -pol], SpinorUbar[p2, -pol], b]) /;
		Mass[p1] == 0 && Mass[p2] == 0,

	Literal[NonCommutativeMultiply[a___, SpinorUbar[p1_, pol1___],
	SpinorU[p2_, pol2___], b___]] :> SpinorT[p1, p2] *
		NonCommutativeMultiply[a] NonCommutativeMultiply[b] /; 
		Mass[p1] == 0 && Mass[p2] == 0 && 
		((pol1 == Left && pol2 == Right) || 
		(pol1 == Left && {pol2} === {}) ||
		({pol1} === {} && pol2 == Right)),

	Literal[NonCommutativeMultiply[a___, SpinorUbar[p1_, pol1___],
	SpinorU[p2_, pol2___], b___]] :> SpinorS[p1, p2] *
		NonCommutativeMultiply[a] NonCommutativeMultiply[b] /; 
		Mass[p1] == 0 && Mass[p2] == 0 && 
		((pol1 == Right && pol2 == Left) || 
		(pol1 == Right && {pol2} === {}) ||
		({pol1} === {} && pol2 == Left)),

(* the below rule is temporary until I find something better *)
(* the rule also appears in STToTraces *)
SpinorS[p1_, p2_]^n1_. * SpinorT[p1_, p2_]^n2_. :>
	(-2 DotProduct[p1,p2])^n1 SpinorT[p1, p2]^(n2-n1) /; (n1>0 && n2>=n1),


SpinorS[p1_, p2_]^n1_. * SpinorT[p1_, p2_]^n2_. :>
	(-2 DotProduct[p1,p2])^n2 SpinorS[p1, p2]^(n1-n2) /; (n1>=n2 && n2>0),

(* make sure these rules come last *)
Eps[a_, b_, c_, d_] :> 
	- I/4 tr[NonCommutativeExpand[NonCommutativeMultiply[
	DiracGamma[a], DiracGamma[b], DiracGamma[c], DiracGamma[d],
	HelicityProjection[Right] - HelicityProjection[Left]]]],

Literal[tr[NonCommutativeMultiply[a___, b_, c___]]] :>
	tr[NonCommutativeExpand[NonCommutativeMultiply[a, Slash[b], c]]] /;
	Mass[b] === 0

}

(* ************************************************************* *)

(* rules about trace, assumes SpaceTimeDimension is 4 *)

tr[] = 4

tr[a_Plus] := tr /@ a

tr[a_ b_] := a tr[b] /; 
	NumberQ[a] || Head[a] === SpinorS || Head[a] === SpinorT

tr[SpinorUbar[p1_, pol_] SpinorU[p2_, -pol_]] := 
	4 NonCommutativeMultiply[SpinorUbar[p2, pol], SpinorU[p1, -pol]]

tr[a_] := 4 a /;
	FreeQ[a, NonCommutativeMultiply] && FreeQ[a, SpinorU] &&
	FreeQ[a, SpinorUbar]

(* ************************************************************* *)


CEvaluatePhaseSpaceIntegrate[expr:PhaseSpaceIntegrate[__], opts___] := 
	cEvaluatePhaseSpaceIntegrate1[
		chooseLightlikeVectorNotCollinearWith[expr], opts]


cEvaluatePhaseSpaceIntegrate1[PhaseSpaceIntegrate[
	integrand_, initmomrules_List, decayrules_List, integrals_List,
	extmom_List, opts___], opts1___] :=
		Block[{man = Mandelstam/.{opts}/.
			Options[CrossSection], 
		sumoverpols =SumOverPolarizations/.{opts1}/.
			Options[CEvaluatePhaseSpaceIntegrate],
		inputfile = Flatten[{InputFile/.{opts1}/.
			Options[CEvaluatePhaseSpaceIntegrate]}],
		outputfile = Flatten[{OutputFile/.{opts1}/.
			Options[CEvaluatePhaseSpaceIntegrate]}],
		dofactor = DoFactor/.{opts1}/.
			Options[CEvaluatePhaseSpaceIntegrate],
		integrandname = ToExpression[
			IntegrandName/.{opts1}/.
			Options[CEvaluatePhaseSpaceIntegrate]],
		couplings = CouplingConstant/.{opts1}/.
			Options[CEvaluatePhaseSpaceIntegrate],
		manrules, dplist, dprules, tmp, tmp1, integrand1, integrand2,
		xrules, integrals1, decayrules1 = Hold[decayrules],
		flatextmom = Flatten[extmom], dpshtcutrules, complexrules = {},
		psumexplicit = Plus @@ (#[[2]]& /@ initmomrules),
		p0symlist = #[[1]]& /@ initmomrules, a, numdim, initmom,
		limits = "", angledeclarations = "", anglefndeclarations = "",
		declarations = "static double ", fvdeclarations = "fv ",
		dotproducts = "", decays = "",
		intermedmom, decaysaslist, lightlikevectorrules, stdefns = "",
		compdeclarations = "complex ", returnstatement, angles = "",
		xs = ""},

(* declare the integrand name *)
	declarations = addvar[declarations, {integrandname}];

(* construct the return statement *)
	returnstatement = StringJoin[
		ToString[StringForm["return ``;", integrandname]], "\n"];

(* do replacements in integrand and decays for LightlikeVectorDecayedFrom *)

	tmp = Union[Cases[Level[integrand, {0,Infinity}],
			LightlikeVectorDecayedFrom[_, _]]];
	lightlikevectorrules =
	Apply[Rule, Transpose[{tmp, Table[Unique["r"], {Length[tmp]}]}], {1}];

(* work out the variables 'limits' and 'numdim' for Splicing *)

		numdim = Length[integrals];
		For[a = 1, a <= numdim, a++, 
			limits = StringJoin[limits,
			ToString[StringForm["xl[`1`] = `2`;\txu[`1`] = `3`;",
			a,
			CForm[N[#2]], CForm[N[#3]]]& @@ integrals[[a]]],"\n"]
			];

(* set up various other variables, check if I need 'static' or
'register' here *) 

(* deal with the initial momenta *)

		initmom = StringJoin @@ (StringJoin[ToString[
			StringForm["fv `` = new_fv(``, ``, ``, ``);",
			#[[1]],
			CForm[N[#[[2]][[1]]]], CForm[N[#[[2]][[2]]]],
			CForm[N[#[[2]][[3]]]], CForm[N[#[[2]][[4]]]]
			]], "\n"]& /@ initmomrules);

(* deal with a special case first, where something like a
	p1+p2 occurs in the first decay in a cross-section
	calculation. Actually this code will even deal with 
	p1+p2+p3+... for arbitrary number of initial particles!
	Wow, great stuff!!! *)

		If[FreeQ[decayrules1, Plus @@ p0symlist] ||
			Length[p0symlist] == 1,
			Null,
			tmp = Unique["psum"];
			initmom = StringJoin[initmom, ToString[
				StringForm["fv `` = new_fv(``, ``, ``, ``);",
				tmp,
			CForm[N[psumexplicit[[1]]]],
			CForm[N[psumexplicit[[2]]]],
			CForm[N[psumexplicit[[3]]]],
			CForm[N[psumexplicit[[4]]]]
			]], "\n"];
			If[MatchQ[psumexplicit, {0,0,0,_}],
				a = {Momentum[Plus @@ p0symlist] -> {0,0,1}},
				a = {}];
			decayrules1 = decayrules1 /. a /. 
				(Plus @@ p0symlist) -> tmp
			];

(* get declarations inc the intermediate momenta.  Add in a zaxis declaration
	in case need it in the C program.  Add in the integration variables.
	Also have to get the variables defining the dotproducts later *)
	(* decaysaslist gets a list of the form:
	{ {{a1,a2},Decay[...]},{{a3,a4},Decay[...]} ...} *)

		decaysaslist = Drop[Flatten[
				Apply[List, decayrules1, {0, 3}], 2], -1] /.
				lightlikevectorrules;
		intermedmom = Complement[Union @@ Cases[
				Flatten[decaysaslist,1], {_,_}], flatextmom];
		fvdeclarations = addvar[fvdeclarations,
				Join[flatextmom, intermedmom]];
		angledeclarations = StringJoin @@ (StringJoin[ToString[
				StringForm["double `` = ``; ", #[[1]],
				#[[2]]]], "\n"]& /@
				Transpose[{First /@ integrals,
				Table[ToString[StringForm["x[``]",tmp1]],
				{tmp1, Length[integrals]}]}]);
		anglefndeclarations = StringJoin @@ 
				Drop[Characters[StringJoin @@ (ToString[
				StringForm["double ``, ", #]]& /@
				First /@ integrals)], -2];
		angles = StringJoin @@ 
				Drop[Characters[StringJoin @@ (ToString[
				StringForm["``, ", #]]& /@
				First /@ integrals)], -2];
		xs = StringJoin @@ 
				Drop[Characters[StringJoin @@ 
				(ToString[StringForm["x[``], ", #]]& /@ 
				Range[Length[integrals]])], -2];


(* pick out the dotproducts that need to be explicitly written out.
	This has to be treated differently to Evaluate because there
	if a dotproduct didn't occur it was never calculated since
	they were SetDelayed.  Here we don't want to calculate
	where we don't need to!! *)
		
	(* before play around with the integrand make sure all
		Mandelstam-type variables are gotten rid of so... 
		(the code is copied from Evaluate) and also set
		up dotproducts for them *)

		dplist = (Append[#,Unique["dp"]])& /@ 
				Combinations[Join[flatextmom, p0symlist], 2];
		If[man === {} ||
		(manrules = Select[{DotProduct[#[[1]], #[[2]]], #[[3]]}& /@ 
		dplist, (Head[First[#]] =!= DotProduct)&]) === {},
		manrules = {},
	(* note Reduce returns 'true' if manrules is {}. Note that headers
		are different depending on the number of solutions that
		Reduce finds, hence the And->List, Or->List and List->And *)
		manrules = Flatten[{Reduce[Apply[Equal, manrules, 1], man] /.
				And -> List /. Or -> List}];
		manrules = ToRules[Union[Select[manrules,
				MemberQ[man, First[#]]&]] /. List -> And];
		Scan[(declarations = addvar[declarations, {#[[3]]}];
			dotproducts = StringJoin[dotproducts,
				ToString[
				StringForm["`` = dotproduct( ``, `` );",
				#[[3]],
				CForm[#[[1]]], CForm[#[[2]]]]],"\n"])&,
				Select[dplist, (!FreeQ[manrules, #[[3]]])&]]];
		

	(* set up short cut rules for dot products i.e for those external
		dot products which have some kind of symmetry.  Needs to
		be done before replacements *)

		dpshtcutrules = (DotProduct[#[[1]],#[[2]]] ->
		(Mass[Parent[#[[1]]]]^2 -Mass[#[[1]]]^2-Mass[#[[2]]]^2)/2)& /@
			Select[extmom, (MatchQ[#,{_,_}] && 
				(Symmetry[Parent[#[[1]]]]===Spherical ||
				Symmetry[Parent[#[[1]]]]===Cylindrical))&];

	(* apply rules *)

		integrand1 = integrand /. lightlikevectorrules /.
				manrules /. dpshtcutrules;

	(* get rid of DotProduct[p, Opposite[p]], I assume that any
occurrence of Opposite will be such that p is massless *)

		integrand1 = integrand1 /. 
			DotProduct[p_, Opposite[p_]] :> 2 Energy[p]^2;

	(* search the integrand squared for DotProducts, get a list
		of distinct ones (ones that weren't contained in
		Mandelstam variables anyway) *)

		a = Apply[List, Union[Cases[
			Level[integrand1, {0,Infinity}], _DotProduct]], {1}];

	(* now put in declarations set up the dotproducts, including
		doing the replacements in the integrand *)

		Scan[	(tmp = Unique["dp"];
			declarations = addvar[declarations, {tmp}];
			dotproducts = StringJoin[dotproducts,
				ToString[
				StringForm["`` = dotproduct( ``, `` );",
				tmp, CForm[#[[1]]], CForm[#[[2]]]]],"\n"];
			integrand1 = integrand1 /.((DotProduct @@ #) -> tmp)
			)&, a];

	(* do replacements for S's and T's (if they exist) *)

		a = Union[Cases[Level[integrand1, {0,Infinity}], _SpinorS]];
		If[a =!= {},
			a = Transpose[{a,
				tmp = Table[Unique["s"], {Length[a]}]}];
			compdeclarations = addvar[compdeclarations, tmp];
			complexrules = Apply[Rule, a, {1}];
			Scan[(stdefns =
				StringJoin[stdefns, ToString[StringForm[
				"`` = SpinorS( ``, `` );", #[[2]],
				CForm[#[[1]][[1]]],
				CForm[#[[1]][[2]]]]], "\n"])&, a];,
			Null];
		a = Union[Cases[Level[integrand1, {0,Infinity}], _SpinorT]];
		If[a =!= {},
			a = Transpose[{a,
				tmp = Table[Unique["t"], {Length[a]}]}];
			compdeclarations = addvar[compdeclarations, tmp];
			complexrules = Join[complexrules, Apply[Rule, a, {1}]];
			Scan[(stdefns = 
				StringJoin[stdefns, ToString[StringForm[
				"`` = SpinorT( ``, `` );", #[[2]],
				CForm[#[[1]][[1]]],
				CForm[#[[1]][[2]]]]], "\n"])&, a];,
			Null];

(* and so finally we come to those nasty decays we have to deal with *)
		

	(* i think i am going to have to take each decay completely apart
		and then put it back together again.  Do this by action
		on decaysaslist *)

		Scan[(tmp = Flatten[# /. Decay -> List, 2];
		decays = StringJoin[decays,
		StringJoin[ToString[StringForm[
		"decay( ``, ``, ``, ``, ``, &``, &``, `` );", tmp[[3]],
		N[tmp[[4]]], N[tmp[[5]]],
		CForm[N[tmp[[6]]]], CForm[N[tmp[[7]]]], tmp[[1]], tmp[[2]],
		If[Length[tmp] < 8, 0,
			Switch[tmp[[8]],
	(* i don't like the next line because the user may have chosen
a different direction.  what i have done is to assume that if there is
a list there it is {0,0,1} *)
				Rule[_, _List], 0,
				Rule[_, Momentum[_]],
					ToString[StringForm[
					"(double *) FV(``)",
					tmp[[8]][[2,1]]]]]],
		]], "\n"]])&, decaysaslist];

(* try to do some factoring on the integrand *)

	Print["starting to factor and splice the integrand"];

	integrand2 = If[Not[sumoverpols] &&
		Head[integrand1] === List &&
		Length[integrand1] > 1 &&
		Length[MasslessInAndOutList] > 0,

			tmp = preparepols[
				N[ComplexForm[
				factor[integrand1, dofactor, couplings]]] /.
					complexrules,
				integrandname];
			compdeclarations = addvar[compdeclarations, tmp[[2]]];
			declarations = addvar[declarations, tmp[[3]]];
			tmp[[1]],

			tmp1 = breakexpr[N[ComplexForm[
			If[Head[
			tmp = factor[integrand1, dofactor, couplings]] === 
					List,
				factor[Plus @@ tmp, dofactor, couplings],
				tmp]]] /. complexrules, integrandname];
			compdeclarations = addvar[compdeclarations,
						Flatten[Take[tmp1, {-2}]]];
			declarations = addvar[declarations, Last[tmp1]];
			StringJoin @@ (StringJoin[ToString[
				StringForm["`` = ``;",#[[2]],CForm[#[[1]]]]],
				"\n", " ", "\n"]& /@ Drop[tmp1, -2])

		];

(* Just making sure declarations ends in a ';' *)

		If[declarations === "static double ", declarations = "",
				declarations = StringJoin[StringJoin @@
				Drop[Characters[declarations], -2],";"]];
		If[fvdeclarations === "fv ", fvdeclarations = "",
				fvdeclarations = StringJoin[StringJoin @@
				Drop[Characters[fvdeclarations], -2],";"]];
		If[compdeclarations === "complex ", compdeclarations = "",
				compdeclarations = StringJoin[StringJoin @@
				Drop[Characters[compdeclarations], -2],";"]];

If[Length[outputfile] == Length[inputfile] && outputfile =!= {""},
	Scan[Splice[#[[1]], #[[2]], FormatType -> OutputForm]&,
		Transpose[{inputfile, outputfile}]],
	Scan[Splice[#, FormatType -> OutputForm]&, inputfile]];


(*

(* And now to Splice... *)
		If[outputfile == "",
			Splice[inputfile, FormatType -> OutputForm],
			Splice[inputfile, outputfile, FormatType -> OutputForm]
		];

*)
		Print["Spliced !!!!  Go ahead and compile it dude!!"];


]

(* ************************************************************* *)

(* a function to append variable names to stringed declarations *)

addvar[string_, list_List] := Block[{newstring = string},
	Scan[(newstring = StringJoin[newstring,
	ToString[StringForm["``, ", #]]])&, list]; newstring]

(* ************************************************************* *)

(* a function that takes a long expression and breaks it up into
something more reasonable.  It takes an expression and its name and
returns a list that contains all the expressions paired with the
variables defining them and two lists of variables to be declared -
the first list is complexes and the second is for doubles *)

MaximumTolerableDepth = 5;

done[expr_] := Block[{head = Head[expr]},
	((Depth[expr] <= MaximumTolerableDepth && head =!= Squared) || 
	(head =!= ComplexTimes && head =!= ComplexPlus &&
	head =!= Times && head =!= Plus &&
	head =!= TimesByDouble && head =!= Squared && 
	Depth[expr] > MaximumTolerableDepth)) &&
	If[head === ComplexPlus, Length[expr] < 15, True]]

(* note this should come before the statement marked  *1*  *)
breakexpr[expr_ComplexPlus, nameofexpr_] :=
	Block[
	{expr1 = expr, newvars, totalexpr, compnewvars, doubnewvars,
	return, listofexprs, firstfew},
	firstfew = Take[List @@ expr1, 12];
	newvars = Table[Unique["expr"], {13}];
	totalexpr = ComplexPlus[Sequence @@ newvars, END];
	return = Apply[breakexpr, 
		Append[Transpose[{firstfew, Drop[newvars, -1]}],
			{Drop[expr1, 12], Last[newvars]}], {1}];
	compnewvars = Flatten[{newvars, Take[#, {-2}]& /@ return}];
	doubnewvars = Flatten[Last[#]& /@ return];
	listofexprs = Append[Flatten[Drop[#, -2]& /@ return, 1],
			{totalexpr, nameofexpr}];
	{Sequence @@ listofexprs, compnewvars, doubnewvars}
	] /;
	Length[expr] >= 15
	
(*  *1*  *)
breakexpr[expr_, nameofexpr_] := 
	Block[
	{expr1 = expr, list, newvars, totalexpr, return,
	listofexprs, compnewvars, doubnewvars, head = Head[expr]},
	Which[
	SameQ[head, Plus],
		Null,
	SameQ[head, ComplexTimes] || SameQ[head, ComplexPlus],
		expr1 = Drop[expr1, -1]
	];
	newvars = Table[Unique["expr"], {Length[expr1]}];
	list = Transpose[{List @@ expr1, newvars}];
	return = Apply[breakexpr, list, {1}];
	Which[
	SameQ[head, Plus], 
		totalexpr = head @@ newvars;
		compnewvars = Flatten[{Take[#, {-2}]& /@ return}];
		doubnewvars = Flatten[{newvars, Last[#]& /@ return}],
	SameQ[head, ComplexTimes] || SameQ[head, ComplexPlus],
		totalexpr = Append[head @@ newvars, END];
		compnewvars = Flatten[{newvars,Take[#, {-2}]& /@ return}];
		doubnewvars = Flatten[{Last[#]& /@ return}]
	];
	listofexprs = Append[Flatten[Drop[#, -2]& /@ return, 1],
			{totalexpr, nameofexpr}];
	{Sequence @@ listofexprs, compnewvars, doubnewvars}
	] /; 
	(Head[expr] === ComplexTimes || Head[expr] === ComplexPlus ||
	Head[expr] === Plus) &&
	Depth[expr] > MaximumTolerableDepth

breakexpr[expr_Times, nameofexpr_] :=
	Block[{list, list1, simple, complicated, newvars, tmp, return,
	totalexpr, compnewvars, doubnewvars, listofexprs},
	list = List @@ expr;
	simple = Select[list, (Depth[#]<3)&];
	tmp = Unique["expr"];
	complicated = Complement[list, simple];
	newvars = Table[Unique["expr"], {Length[complicated]}];
	list1 = Transpose[{List @@ complicated, newvars}];
	return = Apply[breakexpr, list1, {1}];
	totalexpr = Times @@ (Flatten[{tmp, newvars}]);
	compnewvars = Flatten[{Take[#, {-2}]& /@ return}];
	doubnewvars = Flatten[{tmp, newvars, Last[#]& /@ return}];
	listofexprs = Prepend[Flatten[Drop[#, -2]& /@ return, 1],
			{Times @@ simple, tmp}];
	AppendTo[listofexprs, {totalexpr, nameofexpr}];
	{Sequence @@ listofexprs, compnewvars, doubnewvars}
	] /;
	Depth[expr] > MaximumTolerableDepth


breakexpr[expr_Squared, nameofexpr_] :=
	Block[
	{tmp = Unique["expr"], return, totalexpr, compnewvars, doubnewvars},
	return = breakexpr[First[expr], tmp];
	If[ComplexQ[First[expr]],
		totalexpr = Squared[tmp];
		compnewvars = Prepend[Flatten[Take[return, {-2}]], tmp];
		doubnewvars = Last[return],
		totalexpr = Squared[ConstructComplex[tmp, 0.0]];
		compnewvars = Flatten[Take[return, {-2}]];
		doubnewvars = Prepend[Last[return], tmp]
	];
	{Sequence @@ Drop[return, -2],
	{totalexpr, nameofexpr},
	compnewvars, doubnewvars}
	] /;
	Depth[expr] > MaximumTolerableDepth

breakexpr[expr_Squared, nameofexpr_] :=
	Block[
	{tmp},
	If[ComplexQ[First[expr]], {{expr, nameofexpr}, {}, {}},
		tmp = Unique["expr"];
		{{ConstructComplex[First[expr], 0.0], tmp}, 
			{Squared[tmp], nameofexpr}, {tmp}, {}}]
	] /;
	Depth[expr] <= MaximumTolerableDepth

breakexpr[expr_TimesByDouble, nameofexpr_] :=
	Block[
	{comp = Unique["expr"], doub = Unique["expr"], totalexpr,
	compnewvars, doubnewvars, return, listofexprs},
	return = Apply[breakexpr, {{First[expr], comp}, {Last[expr], doub}},
			{1}];
	totalexpr = TimesByDouble[comp, doub];
	compnewvars = Flatten[{comp ,Take[#, {-2}]& /@ return}];
	doubnewvars = Flatten[{doub, Last[#]& /@ return}];
	listofexprs = Append[Flatten[Drop[#, -2]& /@ return, 1],
			{totalexpr, nameofexpr}];
	{Sequence @@ listofexprs, compnewvars, doubnewvars}
	] /;
	Depth[expr] > MaximumTolerableDepth

breakexpr[expr_, nameofexpr_] := 
	{{expr, nameofexpr}, {}, {}} /; done[expr]
	


(* ************************************************************* *)

(* preparepols does some factoring and then returns a string
consisting of the expressions for each individual polarization
combination, and two lists of vars to declare - complexes first and
then doubles.  exprs that are zero are left out of the expression *)


preparepols[expr_List, nameofexpr_] := 
	Block[
	{expr1 = expr, combs, variables, string, dec,
	return, listsofexprs, compnewvars, doubnewvars, comments,
	commentsstrings, intstrings},
	variables = Table[Unique["int"], {Length[expr1]}];
	
	combs = {Insert[#, Left, -1], Insert[#, Right, -1]}& /@ 
		MasslessInAndOutList;
	combs = Flatten[Outer[List, Sequence @@ combs],
		Length[MasslessInAndOutList]-1];
	combs = Transpose[{combs, Transpose[{expr1, variables}]}];
	combs = Cases[combs, {_, {_?(UnsameQ[#, 0.0]&), _}}];
	dec = Last[Last[#]]& /@ combs;
	comments = First[#]& /@ combs;
	return = Apply[breakexpr, Last[#]& /@ combs, {1}];
	compnewvars = Flatten[{Take[#, {-2}]& /@ return}];
	doubnewvars = Flatten[{dec, Last[#]& /@ return}];

	commentsstrings = StringJoin[ToString[StringForm["/* `` */", #]],
				"\n", " ", "\n"]& /@ comments;
	listsofexprs = Drop[#, -2]& /@ return;
	intstrings = Function[x,
			StringJoin @@ 
			(StringJoin[ToString[
			StringForm["`` = ``;", #[[2]], CForm[#[[1]]]]],
				"\n", " ", "\n"]& /@ x)
			] /@ listsofexprs;
	string = StringJoin @@ Flatten[
		{Transpose[{commentsstrings, intstrings}],
		ToString[StringForm["`` = ``;", nameofexpr, Plus @@ dec]]}];

	{string, compnewvars, doubnewvars}
	]
	

(* ************************************************************* *)

(* factor extracts common factors from the elements of a list and then
returns a factored expression.  It will also impose a Factor onto the
argument of a 'Squared'. The second argument (True or False) tells
factor whether to go ahead and factor.  The third argument is a list
of factors that can be pulled out.  If non-empty it overrides a False
in the second argument *)

factor[expr_, False, {}] := expr

factor[expr_, False, x_?(UnsameQ[#, {}]&)] := factor[expr, True, x]

factor[expr_, True, couplings_] := Block[
	{expr1 = expr, list, table, i, tmp},
	expr1 = expr1 /. Complex[a_, b_] :> a + i b;
	list = Union[Select[Level[expr1, {0, Infinity}],
		(NumberQ[#] && !PolynomialQ[#])&]];
	table = Table[Unique["num"], {Length[list]}];

	expr1 = expr1 /. Apply[Rule, Transpose[{list, table}], 1] /. 
		Sqrt[y_] :> tmp[y];
	factor1[expr1, Join[couplings, table, {tmp}]] /.
		Apply[Rule, Transpose[{table, list}], 1] /.
		i -> I /. tmp[y_] :> Sqrt[y]
	]


factor1[expr_List, l_List] := factor1[#, l]&  /@ expr


factor1[expr_?(UnsameQ[Head[#], List]&), l_List] := Block[
	{expr1 = expr},
	If[FreeQ[expr1, Squared],

		factor2[expr1, l],		

		expr1 = Factor[expr1];
		MapAt[factor2[#, l]&, expr1,
			Position[expr1, Squared] /. {{a___, 0} :> {a, 1}}]]]


factor2[expr_, l_List] :=
	Block[
	{expr1 = expr},
	If[expr1 =!= 0, expr1 = FactorTerms[expr1, l], Null];
	expr1 = Collect[#, {SpinorS, SpinorT}]& /@ expr1;
	MapAt[Factor, expr1, Position[expr1,
		Times[__, _SpinorS]]]]


(*
factor1[expr_?(UnsameQ[Head[#], List]&)] := Block[
	{expr1 = expr},
	If[FreeQ[expr1, Squared],
		expr1,
		expr1]]
*)



(* ************************************************************* *)



(* implement some kind of automated process for choosing
LightlikeVectorNotCollinearWith - do the best we can *)

chooseLightlikeVectorNotCollinearWith[expr:PhaseSpaceIntegrate[
	integrand_, initmomrules_List, _, _, extmom_List, ___]] :=
	
	Block[{p0symlist = #[[1]]& /@ initmomrules, tmp, tmp1,
		notcollinearrules = {}, needpard,
		initmomneedpard, extmomneedpard, allotherneedpard,
		zeromassinitmom, zeromassextmom},

	needpard = Union[Cases[Level[integrand, {0,Infinity}],
			LightlikeVectorNotCollinearWith[p_], p]];

	zeromassinitmom = Select[p0symlist, SameQ[Mass[#], 0]&];
	zeromassextmom = Select[Flatten[extmom], SameQ[Mass[#], 0]&];

	(* deal with initial momenta - choose non-collinear one to be
one of the other incoming momenta as long as it is massless *)
	initmomneedpard = Intersection[needpard, p0symlist];
	If[Length[zeromassinitmom] > 1,
		AppendTo[
		notcollinearrules, 
		Rule[LightlikeVectorNotCollinearWith[#], 
		First[Complement[zeromassinitmom, {#}]]]]& /@ initmomneedpard,
	Null];

	(* deal with external momenta - choose the one that is paired
with it in extmom, if it exists and is massless, hopefully they will
never be collinear *)
	extmomneedpard = Complement[needpard, initmomneedpard];
	If[Mass @@ (tmp = 
	Join[	Cases[tmp1 = Level[extmom, {0, Infinity}], 
		{#, p_?(Function[{x}, UnsameQ[Head[x], List]])}, p],
		Cases[tmp1,
		{p_?(Function[{x}, UnsameQ[Head[x], List]]), #}, p]]) === 0,
	AppendTo[notcollinearrules,
		Rule[LightlikeVectorNotCollinearWith[#], First[tmp]]],
	Null]& /@ extmomneedpard;

	(* put the remaining guys into a state where the non-collinear
momentum defaults to an incoming lightlike momentum or one which has
the opposite 3-momentum *)
	allotherneedpard = Complement[needpard,
		Cases[notcollinearrules, 
			Rule[LightlikeVectorNotCollinearWith[p_], _], p]];
	If[zeromassextmom =!= {},
		AppendTo[notcollinearrules,
		Rule[LightlikeVectorNotCollinearWith[#],
			First[zeromassextmom]]]& /@
		Intersection[allotherneedpard, p0symlist],
		AppendTo[notcollinearrules,
		Rule[LightlikeVectorNotCollinearWith[#], Opposite[#]]]& /@
		Intersection[allotherneedpard, p0symlist]
	];
	
	allotherneedpard = Complement[allotherneedpard, initmomneedpard];
	If[zeromassinitmom =!= {},
		AppendTo[notcollinearrules,
		Rule[LightlikeVectorNotCollinearWith[#],
			First[zeromassinitmom]]]& /@ allotherneedpard,
		AppendTo[notcollinearrules,
		Rule[LightlikeVectorNotCollinearWith[#], Opposite[#]]]& /@
		allotherneedpard];
		
	expr /. notcollinearrules

]

(* ************************************************************** *)

(* PrepareMatrixElementForCProgram converts the matrix-element to st
form and does collecting and other simplifications.  The result is
dumped to the named file.  It assumes that all the momenta in the
spinors are massless. *)

PrepareMatrixElementForCProgram[matrixelement_,
	polarizations_,
	contractions_, simplifyrules_, collectconstants_,
	externalmomenta_List, outputfile_String, opts___] := Block[
{man = Mandelstam /. {opts} /. Options[PrepareMatrixElementForCProgram], 
matelem = matrixelement,
pols = Flatten[{polarizations}], conts = Flatten[{contractions}],
simprules = Flatten[{simplifyrules}],
consts = Flatten[{collectconstants}],
momenta = Flatten[RotateLeft[externalmomenta]], ncmrules, polrules, heavymom,
convtomlessrule, ncmpos, ncmlen, massfactor, tmp, replacerules,
nonncm, dprules, dpdec = "static double "},

(* some preparation *)

ncmrules = 
{Literal[NonCommutativeMultiply[
SpinorUbar[p_, pol1___], b___, c_, Slash[p_], d___, SpinorU[q_, pol2___]]] :>
ExpandSlash[Commute[NonCommutativeMultiply[
SpinorUbar[p, pol1], b, c, Slash[p], d, SpinorU[q, pol2]], {c, Slash[p]}]],
Literal[NonCommutativeMultiply[
SpinorUbar[p_, pol1___], b___, Slash[q_], c_, d___, SpinorU[q_, pol2___]]] :>
ExpandSlash[Commute[NonCommutativeMultiply[
SpinorUbar[p, pol1], b, Slash[q], c, d, SpinorU[q, pol2]], {Slash[q], c}]]};

matelem = matelem /. {SpinorV[p_, pol___] :> SpinorU[p, pol] /; Mass[p] == 0, 
		SpinorVbar[p_, pol___] :> SpinorUbar[p, pol] /; Mass[p] == 0};

(* do contractions *)

polrules = Rule[First[#][Last[#]], 0]& /@ pols;
matelem = Contract[ExpandSlash[matelem /. simprules] //. ncmrules, conts];
matelem = matelem /. simprules //. ncmrules /. simprules /. polrules;
If[FreeQ[matelem, Complex], Null, matelem = Expand[matelem/I]];
If[Head[matelem] === List,
	matelem = Collect[#, NonCommutativeMultiply]& /@ matelem,
	matelem = Collect[matelem, NonCommutativeMultiply]];

(* convert a massive momentum via mom-conservation (otherwise pick a
massless one) *)

heavymom = Select[momenta, UnsameQ[Mass[#], 0]&];
If[heavymom === {}, heavymom = First[momenta], heavymom = First[heavymom]];
convtomlessrule = 
	Block[{x = Complement[#, {heavymom}]& /@ externalmomenta, b},
	If[MemberQ[externalmomenta[[1]], heavymom],
		b = Plus @@ x[[2]] - Plus @@ x[[1]],
		b = Plus @@ x[[1]] - Plus @@ x[[2]]];
	{Slash[heavymom] -> ExpandSlash[Slash[b]],
			heavymom[mu_] :> Release[b[mu]]} /. polrules];
matelem = ExpandSlash[matelem /. convtomlessrule /. simprules] //. ncmrules /.
							simprules /. polrules;

Print["I got to here"];
matelem >> z1.m;

(* now for a collection ... any donations ? *)

If[Head[matelem] === List, matelem = Plus @@ matelem, Null];
matelem = Collect[matelem, consts];
matelem = Collect[#,Flatten[{NonCommutativeMultiply, momenta, G}]]& /@ matelem;

(* extractions - any teeth around ? *)

{nonncm, matelem, ncmlen} = disentangle[matelem, momenta];

(* deal with when a G[..] or a p[..] got caught in nonncm. tmp
contains the G's and p's that are in nonncm *)
{nonncm, matelem} = Block[
	{tmp = Times @@ (Flatten[Cases[List @@ nonncm, #]& /@
		(Blank[#]& /@ Flatten[{momenta, G}])])},
	{nonncm/tmp, matelem tmp}];
	

Print["Now I got to here"];
matelem >> z2.m;

(* goto spinor technique form *)

matelem = PolarizationCombinations[matelem Times @@ pols];
Print["Starting to do CTTS"];
matelem = ConvertToST[matelem] /. simprules;

Print["Now I got to here too"];
matelem >> z3.m;

matelem = nonncm matelem;

(* leave all other collecting to CEvaluate and to the intelligence of
the user - for example factors of 1/Masses and Sqrt[3]'s etc *)

(* save the expression and the Replacements *)

OpenWrite[outputfile, FormatType -> TextForm];
dprules = Cases[simprules, Rule[DotProduct[p1_, p2_], x_],
					{DotProduct[p1, p2], x}];
If[man === {}, dprules = (Rule @@ RotateLeft[#])& /@ dprules,
	If[dprules === {}, Null,
	dprules = Flatten[{Reduce[Apply[Equal, dprules, 1], man] /.
				And -> List /. Or -> List}];
	dprules = ToRules[Union[Select[dprules, MemberQ[man, First[#]]&]] /.
				List -> And]]];
dpdec = addvar[dpdec, First /@ dprules];
If[dpdec === "static double ", dpdec = "",
	dpdec = StringJoin[StringJoin @@ Drop[Characters[dpdec], -2],";"]];
Write[outputfile, dpdec];
If[dprules === {}, Null,
Write[outputfile, StringJoin @@ (StringJoin[ToString[
	StringForm["`` = ``;", 	#[[1]], CForm[N[#[[2]] /.
		Power[a_, b_Rational] :> a^N[b] /.
		a_^2 :> SQR[a]]]]], "\n"]& /@ dprules)]];
If[ncmlem === 0, Null,
replacerules = Select[Array[ReplacedExpressionRule, ncmlen], UnsameQ[#, 0]&];
tmp = Count[replacerules, Rule[ComplexReplacedExpression[_], _]];
If[tmp =!= 0,
	Write[outputfile, ToString[StringForm[
		"static complex ComplexReplacedExpression[``];", tmp+1]]],
	Null];
tmp = Count[replacerules, Rule[DoubleReplacedExpression[_], _]];
If[tmp =!= 0,
	Write[outputfile, ToString[StringForm[
		"static double DoubleReplacedExpression[``];", tmp+1]]],
	Null];
Write[outputfile, StringJoin @@ (StringJoin[ToString[
	StringForm["`` = ``;", 	#[[1]], CForm[N[ComplexForm[#[[2]] /.
		Power[a_, b_Rational] :> a^N[b] /.
		a_^2 :> SQR[a]]]]]], "\n"]& /@ replacerules)]];
Write[outputfile, InputForm[matelem]];
Close[outputfile];
matelem]


(* ************************************************************** *)

(* This picks apart common forms of the matrix-element.  WARNING - I
may have forgotten some possibilties.  *)

(* a b c .. NCM[] NCM[]... *)
disentangle[Literal[Times[k1_., k2__NonCommutativeMultiply]], _] :=
	{k1, k2, 0}

(* () ( () NCM NCM ... + () NCM NCM ... + ... ) *)
disentangle[Literal[
	Times[k_., expr:Plus[Times[_., __NonCommutativeMultiply]..]]],
	mom_] :=
	{k,
	Block[{expr1 = List @@ expr, len = Length[expr], real = 0,
	complex = 0},
	While[real + complex < len,
		{expr1[[real+complex+1]], real, complex} =
			extract[expr1[[real+complex+1]], real, complex, mom]];
	Plus @@ expr1],
	Length[expr]}

(* stuff with no NCM's *)
disentangle[expr_, _] := {1, expr, 0} /; FreeQ[expr, NonCommutativeMultiply]

(* ************************************************************** *)

ComplexQ[DoubleReplacedExpression[_]] ^= False

ComplexQ[ComplexReplacedExpression[_]] ^= True

(* This function picks out from an expression the parts that will be
ignored in CTTS, replaces them by ReplacedExpression[number] and sets
up a rule to memorize the replacement. This function will try to do a
little simplification on the expressions *)

extract[expr_NonCommutativeMultiply, real_, complex_, _] :=
(ReplacedExpressionRule[real+complex+1] = 0; {expr, real+1, complex})

extract[expr_Times, real_, complex_, momenta_] := 
Block[{expr1 = expr, constants, ncm, return},
ncm = Select[List @@ expr1, 
	MemberQ[Flatten[{NonCommutativeMultiply, momenta, G}], Head[#]]&];
constants = Times @@ Complement[List @@ expr1, ncm] //.
		a_ b_. + a_ c_. :> a (b+c);
If[ComplexQ[constants],
	ReplacedExpressionRule[real+complex+1] = 
		ComplexReplacedExpression[complex+1] -> constants;
	return = {(Times @@ ncm) ComplexReplacedExpression[complex+1],
			real, complex+1},
	ReplacedExpressionRule[real+complex+1] = 
		DoubleReplacedExpression[real+1] -> constants;
	return = {(Times @@ ncm) DoubleReplacedExpression[real+1],
			real+1, complex}];
return]

(* ************************************************************** *)

End[];

EndPackage[]
